@@ -1,5 +1,6 @@
 /*
- * pgsql, written by Cristian Gafton (c) 2006, rPath, Inc.
+ * pgsql, written by Cristian Gafton (c) 2006-2007, rPath, Inc.
+ *        gafton@rpath.com
  *
  * Based on previous work done by:
  * PyGres, version 2.2 A Python interface for PostgreSQL database. Written by
@@ -28,7 +29,7 @@
  * (darcy@druid.net) subject to the same terms and conditions as above.
  *
  * pgsql has been reworked from PyGreSQL by
- * Cristian Gafton <gafton@rpath.com>, (c) 2006 for rPath, Inc.
+ * Cristian Gafton <gafton@rpath.com>, (c) 2006-2007 for rPath, Inc.
  * subject to the same terms and conditions as above.
 */
 
@@ -56,7 +57,7 @@ static PyObject *Error, *Warning, *InterfaceError,
 	*DatabaseError, *InternalError, *OperationalError, *ProgrammingError,
 	*IntegrityError, *DataError, *NotSupportedError;
 
-static const char *PyPgVersion = "0.9.3";
+static const char *PyPgVersion = "0.9.5";
 static char pg__doc__[] = "Simple Python interface to PostgreSQL DB";
 
 /* default values */
@@ -923,13 +924,15 @@ _pg_fetch_cell(PGresult *result, int row, int col)
     PyObject	*ret;
     char	*cell;
     Oid		oidtype;
-    int		known = 0;	/* keep track if we knew about this datatype */
+    unsigned int cellsize = 0;
+
     if (PQgetisnull(result, row, col)) {
 	Py_INCREF(Py_None);
 	return Py_None;
     }
 
     cell = PQgetvalue(result, row, col);
+    cellsize = PQgetlength(result, row, col);
     switch ((oidtype = PQftype(result, col))) {
 	case BOOLOID:
 	    if (*cell == 't' || *cell == 'T') {
@@ -959,7 +962,7 @@ _pg_fetch_cell(PGresult *result, int row, int col)
 		Py_DECREF(tmp);
 	    }
 	    break;
-	case CASHOID: /* nasty $x,yyy.zzz format */
+	case CASHOID: /* nasty $-x,yyy.zzz format */
 	    {
 		int cashsign = 1;
 		char *cashbuf = malloc(PQgetlength(result, row, col) + 1);
@@ -977,32 +980,48 @@ _pg_fetch_cell(PGresult *result, int row, int col)
 		}
 		cashbuf[k] = 0;
 		ret = PyFloat_FromDouble(strtod(cashbuf, NULL) * cashsign);
+		free(cashbuf);
 	    }
 	    break;
+	case BPCHAROID: /* these are all the string types we usually care about */
+	case VARCHAROID:
+	case NAMEOID:
+	case CHAROID:
+	case TEXTOID:
+	    /* It is assumed that null character terminates these strings */
+	    ret = PyString_FromStringAndSize(cell, cellsize);
+	    break;
 	case BYTEAOID:
 	    if (PQfformat(result, col) == 0) {
 		/* bytea values returned in text mode need to be decoded */
 		char *newStr;
-		int newLen;
+		size_t newLen;
 		newStr = PQunescapeBytea(cell, &newLen);
 		ret = PyString_FromStringAndSize(newStr, newLen);
 		PQfreemem(newStr);
 		break;
 	    }
 	    /* fall-through on else is intentional */
-	case BPCHAROID:
-	case VARCHAROID:
-	case NAMEOID:
-	case CHAROID:
-	case TEXTOID:
-	    known = 1;
-	    /* strings fall through with the others */
 	default:
-	    if (!known)
+	    {
+		void *tmpstr = NULL;
+		int tmplen = 0;
+		int retcode = 0;
 		fprintf(stderr, "WARNING: Unknown datatype %d processed as string. "
 			"Check pg_type.h to decode this OID\n", oidtype);
-	    ret = PyString_FromString(cell);
-	    break;
+		if (!(ret = PyBuffer_New(cellsize)) || ret == PyExc_ValueError) {
+		    PyErr_SetString(PyExc_MemoryError, "could not allocate buffer object for row data");
+		    return NULL;
+		}
+		retcode = PyObject_AsWriteBuffer(ret, &tmpstr, &tmplen);
+		if (retcode < 0 || (tmplen != cellsize)) {
+		    PyErr_SetString(InternalError, "could not convert unknown data field to a buffer object");
+		    Py_DECREF(ret);
+		    return NULL;
+		}
+		memcpy(tmpstr, cell, cellsize);
+		break;
+	    }
     }
     return ret;
 }
@@ -1655,9 +1674,9 @@ pgconnect(pgobject * self, PyObject * args, PyObject * dict)
 				   pgdbname, pguser, pgpasswd);
 
     if (PQstatus(npgobj->cnx) == CONNECTION_BAD) {
+	PyErr_SetString(ProgrammingError, PQerrorMessage(npgobj->cnx));
 	PQfinish(npgobj->cnx);
 	npgobj->cnx = NULL; /*noop*/
-	PyErr_SetString(InternalError, PQerrorMessage(npgobj->cnx));
 	Py_XDECREF(npgobj);
 	return NULL;
     }
@@ -1849,7 +1868,7 @@ static char pg_execute__doc__[] =
 "query(sql[,params]) -- creates a new query object for this connection,"
 " using sql (string) request and optionally positional params.";
 static PyObject *
-pg_execute(pgobject *self, PyObject * args)
+pg_execute(pgobject *self, PyObject *args)
 {
     pgsourceobject	*src;
     PyObject	*ret;
@@ -1887,7 +1906,7 @@ static char pg_escape_string__doc__[] =
 static PyObject *
 pg_escape_string(pgobject *self, PyObject *args) {
     char	*from;		/* our string argument */
-    size_t	from_length;	/* length of string */
+    int		from_length;	/* length of string */
     char	*to = NULL;	/* the result */
     size_t	to_length = 0;	/* length of result */
     PyObject	*ret;		/* string object to return */
@@ -1926,7 +1945,7 @@ static char pg_escape_bytea__doc__[] =
 static PyObject *
 pg_escape_bytea(pgobject *self, PyObject *args) {
     unsigned char	*from;		/* our string argument */
-    size_t		from_length;	/* length of string */
+    int			from_length;	/* length of string */
     unsigned char	*to = NULL;	/* the result */
     size_t		to_length = 0;	/* length of result */
     PyObject		*ret;		/* string object to return */
@@ -1957,7 +1976,7 @@ static PyObject *
 pg_unescape_bytea(pgobject *self, PyObject *args)
 {
     unsigned char	*from;		/* our string argument */
-    size_t		from_length;	/* source length */
+    int			from_length;	/* source length */
     unsigned char	*to = NULL;	/* the result */
     size_t		to_length = 0;	/* length of result string */
     PyObject		*ret;	/* string object to return */
@@ -2021,6 +2040,10 @@ pg_setnotices(pgobject *self, PyObject *args)
 #include "pglarge.c"
 #endif /* LARGE_OBJECTS */
 
+#ifdef DIRECT_ACCESS
+#include "pgcopy.c"
+#endif
+
 /* connection object methods */
 static struct PyMethodDef pgobj_methods[] = {
 	{"source", (PyCFunction) pg_source, METH_VARARGS, pg_source__doc__},
@@ -2041,7 +2064,9 @@ static struct PyMethodDef pgobj_methods[] = {
 	{"getlo",    (PyCFunction) pg_getlo, 1, pg_getlo__doc__},
 	{"loimport", (PyCFunction) pg_loimport, 1, pg_loimport__doc__},
 #endif   /* LARGE_OBJECTS */
-
+#ifdef DIRECT_ACCESS
+	{"bulkload",  (PyCFunction) pg_inserttable, METH_VARARGS, pg_inserttable__doc__},
+#endif
 	{"setnotices", (PyCFunction) pg_setnotices, METH_VARARGS, pg_setnotices__doc__},
 
 	{NULL, NULL}				/* sentinel */
